{ File   : ADBExt.dpr
  Author : Deniz Oezmen
  Created: 2005-04-19
  Changed: 2005-10-08

  Nibiru ADB extractor.
}
program ADBExt;

{$APPTYPE CONSOLE}

uses
  SysUtils;

type
  TADBHeader = packed record
                 ID,                      {666}
                 Unknown1,
                 NumEntries,
                 NumEntriesUsed,
                 Unknown2      : Longword
               end;

  TADBEntry  = packed record
                 Offset: Longword;
                 Name  : array[0..31] of Char;
                 Size  : Longword
               end;

var
  ADBHeader: TADBHeader;
  ADBFile  : file;

function OpenADBFile(var ADBFile: file; ADBName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(ADBFile, ADBName);

  {$I-}
  Reset(ADBFile, 1);
  {$I+}
  OpenADBFile := (IOResult = 0)
end;

function ReadHeader: Boolean;
var
 BytesRead: Longword;
begin
  BlockRead(ADBFile, ADBHeader, SizeOf(TADBHeader), BytesRead);
  ReadHeader := (BytesRead = SizeOf(TADBHeader))
end;

procedure ExtractFiles;
var
  BaseOffs,
  OldOffs,
  i, j    : Longword;
  ADBEntry: TADBEntry;
  OutFile : file of Byte;
  Buffer  : array[0..10239] of Byte;
begin
  BaseOffs := ADBHeader.NumEntries * SizeOf(TADBEntry) + SizeOf(TADBHeader);

  for i := 1 to ADBHeader.NumEntriesUsed do
  begin
    BlockRead(ADBFile, ADBEntry, SizeOf(TADBEntry));
    Writeln('Extracting ', ADBEntry.Name, ' @ ', ADBEntry.Offset, ', size: ',
      ADBEntry.Size);

    OldOffs := FilePos(ADBFile);
    Seek(ADBFile, ADBEntry.Offset + BaseOffs);

    FileMode := fmOpenReadWrite;
    AssignFile(OutFile, ADBEntry.Name);
    Rewrite(OutFile);

    for j := 1 to ADBEntry.Size div SizeOf(Buffer) do
    begin
      BlockRead(ADBFile, Buffer, SizeOf(Buffer));
      BlockWrite(OutFile, Buffer, SizeOf(Buffer))
    end;
    BlockRead(ADBFile, Buffer, ADBEntry.Size mod SizeOf(Buffer));
    BlockWrite(OutFile, Buffer, ADBEntry.Size mod SizeOf(Buffer));

    Close(OutFile);

    Seek(ADBFile, OldOffs)
  end
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  ADBName  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara 
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      ADBName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenADBFile(ADBFile, ADBName) then
        if ReadHeader then
        begin
          Writeln('Extracting from ', ADBName, ' ...');
          ExtractFiles; 
          Close(ADBFile)
        end
        else
          Writeln(ADBName, ' is not a valid ADB file.')
      else
        Writeln(ADBName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount = 1 then
    ParseFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.
