{ File   : GRPExt.dpr
  Author : Deniz Oezmen
  Created: 2005-04-19
  Changed: 2007-08-14

  Nibiru/Black Mirror GRP extractor.
}
program GRPExt;

{$APPTYPE CONSOLE}

uses
  SysUtils, DateUtils, Windows;

type
  TGRPHeader = packed record
    ID            : array[0..15] of Char;
    HeaderLength  : Longword;
    Signature     : Longword;
    Version       : Longword;
    NumEntriesUsed: Longword;
    NumEntries    : Longword;
    DateStamp1    : Longword;
    DateStamp2    : Longword;
  end;

  TGRPEntry = packed record
    Name     : array[0..32] of Char;
    Offset   : Longword;
    Size     : Longword;
    DateStamp: Longword;
    Unknown  : Longword;
  end;

function OpenGRPFile(var GRPFile: file; GRPName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(GRPFile, GRPName);

  {$I-}
  Reset(GRPFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function DecodeString(const Code: string): string;
const
  KeyPhrase: array[0..51] of Char =
    'Vyvojovy tym AGDS varuje: Hackerovani skodi obchodu!';
var
  i: Longint;
begin
  SetLength(Result, Length(Code));

  for i := 0 to Length(Code) do
  begin
    if Code[i + 1] = #000 then
    begin
      SetLength(Result, i);
      Break
    end;

    Result[i + 1] := Char(Ord(Code[i + 1]) xor
      Ord(KeyPhrase[i mod Length(KeyPhrase)]) xor $ff)
  end
end;

function ReadHeader(var GRPFile: file; out GRPHeader: TGRPHeader): Boolean;
var
  BytesRead: Longword;
begin
  BlockRead(GRPFile, GRPHeader, SizeOf(TGRPHeader), BytesRead);
  Result := (BytesRead = SizeOf(TGRPHeader));
  if not Result then
    Exit;

  with GRPHeader do
    Result := ((DecodeString(ID) = 'AGDS group file'#$1a) and
      (HeaderLength = SizeOf(GRPHeader)) and (Signature = $1a03c9e6) and
      (Version = 2))
end;

procedure SetDateStamp(const FileName: string; DateStamp: Longword);
var
  Handle  : Longword;
  SysTime : TSystemTime;
  FileTime: TFileTime;
begin
  with SysTime do
    DecodeDateTime(DateStamp / 86400 + UnixDateDelta, wYear, wMonth, wDay,
    wHour, wMinute, wSecond, wMilliseconds);

  SystemTimeToFileTime(SysTime, FileTime);
  LocalFileTimeToFileTime(FileTime, FileTime);

  Handle := FileOpen(FileName, fmOpenWrite);
  if Handle = INVALID_HANDLE_VALUE then
    Exit;

  SetFileTime(Handle, @FileTime, @FileTime, @FileTime);
  
  FileClose(Handle)
end;

procedure ExtractFiles(var GRPFile: file; const GRPHeader: TGRPHeader);
var
  i, j      : Longword;
  GRPEntries: array of TGRPEntry;
  OutFile   : file;
  OutName   : string;
  Buffer    : array[0..102390] of Byte;
begin
  // load directory
  SetLength(GRPEntries, GRPHeader.NumEntriesUsed);
  {$I-}
  BlockRead(GRPFile, GRPEntries[0], SizeOf(TGRPEntry) * Length(GRPEntries));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  error reading directory');
    Exit
  end;

  // read through all valid entries
  for i := 0 to High(GRPEntries) do
    with GRPEntries[i] do
    begin
      // recover filename
      OutName := DecodeString(Name);

      WriteLn('  ', OutName);

      // create output
      {$I-}
      FileMode := fmOpenWrite;
      AssignFile(OutFile, OutName);
      Rewrite(OutFile, 1);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('  file creation error');
        Continue
      end;

      // copy file data
      {$I-}
      Seek(GRPFile, Offset);
      for j := 1 to Size div SizeOf(Buffer) do
      begin
        BlockRead(GRPFile, Buffer, SizeOf(Buffer));
        BlockWrite(OutFile, Buffer, SizeOf(Buffer))
      end;
      BlockRead(GRPFile, Buffer, Size mod SizeOf(Buffer));
      BlockWrite(OutFile, Buffer, Size mod SizeOf(Buffer));
      {$I+}
      if IOResult <> 0 then
        WriteLn('  extraction error');

      CloseFile(OutFile);

      // finally set original file date
      SetDateStamp(ExpandFileName(OutName), DateStamp)
    end;

  Finalize(GRPEntries);
  SetLength(GRPEntries, 0)
end;

procedure ScanFiles(FileParam: string);
var
  FileName,
  Dir,
  GRPName  : string;
  GRPFile  : file;
  GRPHeader: TGRPHeader;
  SearchRec: TSearchRec;
begin
  // get the directory information from FileParam
  Dir := ExtractFilePath(FileParam);

  // search for all files matching FileParam
  if FindFirst(FileParam, faAnyFile - faDirectory, SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      GRPName := Dir + FileName;

      WriteLn(GRPName);

      if OpenGRPFile(GRPFile, GRPName) then
      begin
        if ReadHeader(GRPFile, GRPHeader) then
          ExtractFiles(GRPFile, GRPHeader)
        else
          WriteLn('  not a valid GRP file')
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [FileParam]));

  SysUtils.FindClose(SearchRec)
end;

begin
  if ParamCount = 1 then
    ScanFiles(ParamStr(1))
  else
    WriteLn('Incorrect number of parameters.')
end.
